﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using UniRx;
using UnityEngine;
using static Magify.MagifyManager;

namespace Magify
{
    internal class PurchaserService : BasePurchaserService, IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(MagifyService.LogScope);

        private readonly MagifySettings _settings;
        private readonly ServicePrefs _prefs;
        private readonly ProductsObtainer _obtainer;

        public bool PurchasingInProgress { get; private set; }
        public IReadOnlyList<InAppProduct> InAppProducts => MagifyManager.InAppProducts;
        public IReadOnlyList<SubscriptionProduct> SubscriptionProducts => MagifyManager.SubscriptionProducts;
        public new IInAppStore InAppStore { get; private set; }

        internal PurchaserService(MagifySettings settings, ServicePrefs prefs, ProductsObtainer obtainer)
        {
            _settings = settings;
            _prefs = prefs;
            _obtainer = obtainer;
        }

        void IDisposable.Dispose()
        {
            if (InAppStore != null)
            {
                InAppStore.OnPurchaseFinished -= ExternalPurchaseFinishedHandler;
                OnConfigLoaded -= FetchDefaultProducts;
            }
        }

        public void SetPurchasingProvider([CanBeNull] IInAppStore inAppStore)
        {
            if (InAppStore != null)
            {
                InAppStore.OnPurchaseFinished -= ExternalPurchaseFinishedHandler;
                OnConfigLoaded -= FetchDefaultProducts;
            }
            InAppStore = inAppStore;
            if (InAppStore != null)
            {
                OnConfigLoaded += FetchDefaultProducts;
                InAppStore.OnPurchaseFinished += ExternalPurchaseFinishedHandler;
                FetchDefaultProducts();
            }
        }

        public async UniTask<ProductPrepareResult> PrepareProductAsync(ProductDef product, CancellationToken cancellationToken)
        {
            var timeout = (float)_settings.ProductFetchTimeout;
            if (timeout <= 0)
            {
                return ProductPrepareResult.Fail(ProductObtainFailReason.Timeout);
            }

            var promise = new UniTaskCompletionSource<ProductObtainFailReason?>();
            InAppStore.OnProductFetched += fetched;
            InAppStore.OnProductFetchFailed += failed;

            try
            {
                InAppStore.LoadProducts(new[] { product });
                var result = await promise.Task
                    .AttachExternalCancellation(cancellationToken)
                    .Timeout(TimeSpan.FromSeconds(timeout));

                return result == null ? ProductPrepareResult.Success() : ProductPrepareResult.Fail(result.Value);
            }
            catch (TimeoutException)
            {
                return ProductPrepareResult.Fail(ProductObtainFailReason.Timeout);
            }
            catch (OperationCanceledException)
            {
                throw;
            }
            catch (Exception e)
            {
                return ProductPrepareResult.Fail(ProductObtainFailReason.Unknown, e.GetType().Name);
            }
            finally
            {
                InAppStore.OnProductFetched -= fetched;
                InAppStore.OnProductFetchFailed -= failed;
            }

            void fetched(string productId)
            {
                promise.TrySetResult(null);
            }

            void failed(string productId, ProductObtainFailReason reason)
            {
                promise.TrySetResult(reason);
            }
        }

        public UniTask<(ProductObtainFailReason? FailReason, PurchaseInfo Info)> PurchaseProductAsync(InAppProduct product)
        {
            return PurchaseProductAsync((ProductDef)product);
        }

        public UniTask<(ProductObtainFailReason? FailReason, PurchaseInfo Info)> PurchaseProductAsync(SubscriptionProduct product)
        {
            return PurchaseProductAsync((ProductDef)product);
        }

        private async UniTask<(ProductObtainFailReason? FailReason, PurchaseInfo Info)> PurchaseProductAsync(ProductDef product)
        {
            _logger.Log($"Obtaining {product.GetType().Name} ({product.Id}): to obtain an in-app product, we need to use IPurchasesProvider");
            var purchasePromise = new UniTaskCompletionSource<(ProductObtainFailReason? FailReason, PurchaseInfo Info)>();
            InAppStore.OnPurchaseFinished += purchased;
            InAppStore.OnPurchaseFailed += failed;
            try
            {
                PurchasingInProgress = true;
                InAppStore.Purchase(product);
                var result = await purchasePromise.Task;
                return result;
            }
            catch (Exception e)
            {
                Debug.LogException(e);
                return (ProductObtainFailReason.Unknown, null);
            }
            finally
            {
                await UniTask.SwitchToMainThread();
                InAppStore.OnPurchaseFinished -= purchased;
                InAppStore.OnPurchaseFailed -= failed;
                PurchasingInProgress = false;
            }

            void purchased(string productId, PurchaseInfo purchaseInfo)
            {
                _prefs.PurchasesCounter.Value++;
                purchasePromise.TrySetResult((null, purchaseInfo));
            }

            void failed(string productId, ProductObtainFailReason reason)
            {
                purchasePromise.TrySetResult((reason, null));
            }
        }

        private void FetchDefaultProducts()
        {
            if (InAppStore == null)
            {
                _logger.Log($"Cannot fetch products: {nameof(InAppStore)} is null");
                return;
            }
            if (MainThreadDispatcher.IsInMainThread)
            {
                fetch();
            }
            else
            {
                switchToMainThreadAndFetch().Forget();
            }
            return;

            async UniTaskVoid switchToMainThreadAndFetch()
            {
                await UniTask.SwitchToMainThread();
                fetch();
            }

            void fetch()
            {
                var products = new List<ProductDef>();
                products.AddRange(InAppProducts);
                products.AddRange(SubscriptionProducts);
                _logger.Log($"Fetching default products from store {InAppStore.GetType()}: {string.Join(", ", products)}");
                InAppStore.LoadProducts(products);
            }
        }

        private void ExternalPurchaseFinishedHandler(string productId, PurchaseInfo purchaseInfo)
        {
            if (PurchasingInProgress) return;
            var product = (ProductDef)InAppProducts.FirstOrDefault(c => c.Id == productId) ?? SubscriptionProducts.FirstOrDefault(c => c.Id == productId);
            switch (product)
            {
                case InAppProduct:
                    _logger.Log($"There is external purchase of {productId} detected");
                    TrackExternalInAppFor(
                        product.Id,
                        purchaseInfo.Price,
                        purchaseInfo.Currency,
                        purchaseInfo.TransactionId,
                        purchaseInfo.PurchaseToken,
                        purchaseInfo.OriginalTransactionId,
                        purchaseInfo.Receipt.Payload);
                    _obtainer.HandleExternalObtain(product, purchaseInfo);
                    break;
                case SubscriptionProduct:
                    _logger.Log($"Handle subscription for {productId}. Tracking will be in {nameof(SubscriptionService)}");
                    _obtainer.HandleExternalObtain(product, purchaseInfo);
                    break;
                default:
                    _logger.LogError($"No information about product {productId} in Magify!");
                    break;
            }
        }
    }
}